#!/tvbin/tivosh
# $Id: httpd-tt.tcl,v 1.51.2.9 2002/06/26 22:30:11 lightn Exp $

if {[pool pool0 size] <= 1458176} {
    puts "Use the 'tivoweb' script to start tivoweb"
    exit
}

setpri fifo 1
catch {EnableTransactionHoldoff true}

set mfsVerboseG 0

source $tcl_library/tv/log.tcl
source $tcl_library/tv/mfslib.tcl
source $tcl_library/tv/dumpobj.tcl

global source_dir
set source_dir [file dirname [info script]]

source $source_dir/httpd-tt.itcl
#namespace import TT_HTTPD::*
source $source_dir/html.itcl
source $source_dir/util.itcl

puts "$TT_HTTPD::STARTUP_MESSAGE"

proc register_module {url sdesc ldesc} {
    global module_list
    if { [string first "." $url] == -1 } {
       lappend module_list [list "$url/" $sdesc $ldesc]
    } else {
       lappend module_list [list $url $sdesc $ldesc]
    }
}

proc make_menu {} {
    global module_list

    # set module_list [linsert $module_list 0 [list "" "Main Menu" ""]]
    # lappend $module_list 0 [list "quit" "Quit" "Terminate $TT_HTTPD::NAME"]

    set TT_HTTPD::OPTIONS_MENU "\n"
    append TT_HTTPD::OPTIONS_MENU [html_link "/" "Main Menu"]
    append TT_HTTPD::OPTIONS_MENU "\n"

    set TT_HTTPD::MAIN_MENU "
<TABLE border=0 cellpadding=0 cellspacing=1>
<TR ALIGN=CENTER><TH COLSPAN=2>$TT_HTTPD::NAME v$TT_HTTPD::VERSION</TD></TR>"
    append TT_HTTPD::MAIN_MENU "\n"

    foreach module $module_list {
        append TT_HTTPD::OPTIONS_MENU [html_link "/[lindex $module 0]" [lindex $module 1]]
        append TT_HTTPD::OPTIONS_MENU "\n"

        append TT_HTTPD::MAIN_MENU [tr "ALIGN=LEFT" [td [html_link "/[lindex $module 0]" [lindex $module 1]]] [td [lindex $module 2]]]
        append TT_HTTPD::MAIN_MENU "\n"
    }

    append TT_HTTPD::OPTIONS_MENU [html_link "/restart" "Restart"]

    append TT_HTTPD::MAIN_MENU [tr "ALIGN=CENTER" [td "COLSPAN=2" ""]]
    append TT_HTTPD::MAIN_MENU "\n"
    append TT_HTTPD::MAIN_MENU [tr "ALIGN=LEFT" [td [html_link "/restart" "Restart"]] [td "Reload or Quit $TT_HTTPD::NAME"]]
    append TT_HTTPD::MAIN_MENU "\n"
    append TT_HTTPD::MAIN_MENU [html_table_end]

}

proc action_restart {chan path env} {
    global reload

    if {[string index $path 0] == "/"} {
        set path [string range $path 1 end]
    }

    if { $path == "" } {
       puts $chan [html_start "Restart"]
       puts $chan [html_table_start "" "" "ALIGN=TOP"]
       puts $chan [tr "" [th "" "Restart"]]
       puts $chan [tr "" [td [html_link "/restart/1" "Quick Reload"]]]
       puts $chan [tr "" [td [html_link "/restart/0" "Full Reload"]]]
       puts $chan [tr "" [td [html_link "/quit" "Quit"]]]
       puts -nonewline $chan [html_table_end]
       puts $chan [html_end]
    } else {
       set reload $path
       # puts "Loading modules..."
       puts $chan [html_start "Reload"]
       puts $chan "<B>Reload Complete</B>"
       puts $chan [html_end]
    }
}

proc action_quit {chan dummy env} {
    set ::reload "quit"
    puts $chan [html_start ""]
    puts $chan "Server has terminated."
    puts $chan [html_end]
}


proc action_ {chan dummy env} {
    puts $chan [html_start ""]
    puts $chan "$TT_HTTPD::MAIN_MENU"
    puts $chan [html_end]
}

proc action_robots {chan dummy env} {
    puts $chan "User-agent: *"
    puts $chan "Disallow: /"
}

proc register_content_handler {suffix mimetype directory binary function} {
    global content_suffix_list
    global content_handler_list

    lappend content_suffix_list $suffix
    lappend content_handler_list [list $mimetype $directory $binary $function]
}

proc serve_file {chan filename head_req last_modified} {
    global db
    global source_dir
    global content_suffix_list
    global content_handler_list

    set suffix [file extension $filename]
    if {[string index $suffix 0] == "."} {
        set suffix [string range $suffix 1 end]
    }
    set index [lsearch $content_suffix_list $suffix]
    set clist [lindex $content_handler_list $index]
    set mimetype [lindex $clist 0]
    set directory [lindex $clist 1]
    set binary [lindex $clist 2]

    set fd ""
    set filedata ""
    catch { set fd [open "$directory/$filename" "r"] }
    if { $fd == "" } {
       if { $last_modified != 0 } {
          set filedata "cached"
          set moddate $last_modified
       }
    } else {
       set moddate [file mtime "$directory/$filename"]
    }
    if { ($fd != "") || ($filedata != "") } {
        if { $head_req == 1 } {
            print_html_header_200 $chan $mimetype $moddate
        } elseif { $last_modified == $moddate } {
            print_html_header_304 $chan
        } else {
            print_html_header_200 $chan $mimetype $moddate
            if { $binary == 1 } {
               fconfigure $chan -translation binary
            }
            if { $fd != "" } {
               if { $binary == 1 } {
                  fconfigure $fd -translation binary
               }
               fcopy $fd $chan
            } else {
               puts -nonewline $chan $filedata
            }
        }
        if { $fd != "" } {
           close $fd
        }
    } else {
        print_html_header_404 $chan
    }
}

proc serve_image {chan imagename head_req last_modified} {
    global db
    global source_dir

    set imagename [file rootname $imagename]
    set fd ""
    set imagedata ""
    catch { set fd [open "$source_dir/images/$imagename.png" "r"] }
    if { $fd == "" } {
       if { $last_modified != 0 } {
          set imagedata "cached"
          set moddate $last_modified
       } else {
       RetryTransaction {
          if {[catch {set obj [db $db open "/Resource/Image/$imagename"]}] != 1} {
             if {[catch {set imageid [dbobj $obj get "File"]}] != 1} {
                set imagedata [ mfs get $imageid "tyFile" ]
                if {[catch {set moddate [mfs moddate $imageid]}]} {
                   set moddate ""
                }
             }
          }
       }
       }
    } else {
       set moddate [file mtime "$source_dir/images/$imagename.png"]
    }
    if { ($fd != "") || ($imagedata != "") } {
        if { $head_req == 1 } {
            print_html_header_200 $chan "image/png" $moddate
        } elseif { $last_modified == $moddate } {
            print_html_header_304 $chan
        } else {
            print_html_header_200 $chan "image/png" $moddate
            fconfigure $chan -translation binary
            if { $fd != "" } {
               fconfigure $fd -translation binary
               fcopy $fd $chan
            } else {
               puts -nonewline $chan $imagedata
            }
        }
        if { $fd != "" } {
           close $fd
        }
    } else {
        print_html_header_404 $chan
    }
}

proc decode_upload {chan} {
    global source_dir
    global content_boundary

    #get the first line of the body of the http body. it
    #should be the boundary text, the start of the first part
    gets $chan content_line
    if {$content_line == "--$content_boundary"} {
       while {1} {
          #this is the header info we really need:
          #filename is the name of the file uploaded
          #name is the name of the field in the HTML form
          set filename ""
          set name ""
                        
          #loop though the header info of the form part
          #there is a blank line between the hearder and
          #the actual content, break when you hit that blank line
          while {[gets $chan header_line] >= 0} {
             if {$header_line == ""} { break }

             #check the header for a filename
             #thanks to davew on #tcl for the regexp help.
             #(simple regular expressions, Jake, you should know this. Hehe)
             if {[regexp -nocase {.*filename="([^"]*)\"} $header_line junk filename] == 1} {
                #some browsers give the full path
                #discard path and give just the filename.
                regexp -nocase {[\\/]?([^\\/]*)$} $filename dummy filename
             }
             regexp -nocase {^Content-Disposition:.*\ name="([^"]*)\"} $header_line junk name
          }
          puts "done with headers"
          #get the data. if there's a filename, output the data to the file
          #if there is no filename, just save a variable named "value"
          set filehandle ""
          set value ""

          if {$filename != ""} { 
             set filehandle [open "$source_dir/uploads/$filename" w]
             fconfigure $filehandle -translation binary 
          }

          #for some reason, the chan must buffer or the file corrupts
          fconfigure $chan -buffering full -translation binary

          #create a fifo big enough to hold the boundary string.
          #we will read the data from the client into this fifo and
          #out into the file.  If the fifo is filled with the boundary
          #text exactly, we've reached the EOF.  Load the fifo with
          #the first set of characters
          set fifo_length [string length "\r\n--$content_boundary"]
          set fifo_buffer [read $chan $fifo_length]

          while {$fifo_buffer != "\r\n--$content_boundary"} {
             #get the first byte out of the fifo and write it to
             #the file, or to a variable, depending on what's going on
             set outputByte [string index $fifo_buffer 0]
             if {$filehandle != "" } {
                puts -nonewline $filehandle $outputByte
             } else {
                append value $outputByte
             }
   
             #get the next byte from the client, and add it to the
             #end of the fifo, drop off the head of the fifo we
             #just used a second ago.
             set nextByte [read $chan 1]
             set fifo_buffer [string range "$fifo_buffer$nextByte" 1 end]
          }
                        
          #Generate the form data to put into post_data
          #this gets put into the env variable and passed to
          #the action.  httpMapReply will encode the data
          #into URL form, since thats what post_data expects
          #multipart forms don't encode the data that way.
          set key [httpMapReply $name]
          if {$filehandle != ""} { 
             close $filehandle
             set val [httpMapReply $filename]
          } else {
             set val [httpMapReply $value]
          }

          #If post data is not empty, put the & which separates
          #the different fields
          if {$post_data != ""} { append post_data "\&" }
          append post_data "$key=$val"

          #get the next two bytes. (Reuse nextByte)
          set nextByte [read $chan 2]

          #reset the channel back to what it was before
          fconfigure $chan -buffering none -translation auto

          #if there's two more --'s after the boundary
          #we're at the end of all the parts.  If there's
          #a new line, there's more to come.  If there's
          #something else, there's a problem.
          if {$nextByte == "--"} { break }
          if {$nextByte != "\r\n"} {
             puts "ERROR! Expected new line"
          }
       }
       #reconfigure the channel back to what it was before all this.
       fconfigure $chan -blocking 1 -translation auto
    }
}

proc session {chan addr port} {
    #global db
    #global source_dir

    set auth 0
    set head_req 0
    set post_req 0
    set post_data ""
    set if_modified_since ""
    set last_modified 0
    set content_length 0

    # added by JAKE -
    # http_upload - flag indicates if body has multipart form to parse
    # content_boundary - holds the boundary text between the parts 
    set http_upload 0
    set content_boundary ""

    fconfigure $chan -buffering none -blocking 1
    while {[gets $chan line] >= 0} {
        if {$line == ""} {
           if { $post_req == 1 } {
	      #added by JAKE - check to see if there is a multipart form
	      #if so, parse it, if not, handle as TivoWeb always did.
	      if {$http_upload == 1} {
                 decode_upload $chan
	      } else {
		 #Just another POST (standard httpd-tt code)	
                 append post_data [read $chan $content_length]
	      }
           }
           break
        }

        #if {[regexp -nocase {^kill} $line]  == 1} {
        #    set quit "puts killed"
        #    return
        #}
        if {[regexp -nocase {^get +([^\?]*)\?*(.*) +http/[0-9]+\.[0-9]+.?$} $line dummy path post_data] == 1} {
            continue
        }
        if {[regexp -nocase {^head +([^\?]*)\?*(.*) +http/[0-9]+\.[0-9]+.?$} $line dummy path post_data] == 1} {
        #    set quit 1
        #    return
            set head_req 1
            continue
        }
        if {[regexp -nocase {^post +([^\?]*)\?*(.*) +http/[0-9]+\.[0-9]+.?$} $line dummy path post_data] == 1} {
            set post_req 1
            continue
        }
        if {[regexp -nocase {^Content-length: ([0-9]+)$} $line dummy content_length] == 1} {
            continue
        }

	#Content-type added by JAKE - The content-type header is present
	#and has multipart/form data.  set the http_upload flag, and
	#get the boundary string.
	if {[regexp -nocase {^Content-Type: multipart/form-data; boundary=(.*)$} $line junk content_boundary] == 1} {
	    set http_upload 1
	    continue
        }

        if {$::userpass != ""} {
            if {[regexp -nocase {^Authorization: +Basic +([A-Za-z0-9+/=]+)$} $line dummy authcode] == 1} {
                set authdecode [base64dec $authcode]
                if { $authdecode == $::userpass } {
                   set auth 1
                }
	        continue
            }
        }
        if {[regexp -nocase {^If-Modified-Since: ([^;]*).*$} $line dummy if_modified_since] == 1} {
            set last_modified [clock scan $if_modified_since]
            continue
        }
    }

    if {$::userpass != "" && $auth == 0} {
        print_html_header_401 $chan
        catch {flush $chan}
        catch {close $chan}
        return
    }

    set path [url_decode $path]
    if {[regexp -nocase {/([^/]+)\.([A-Z0-9]+)$} $path dummy filename suffix] == 1} {
        set index [lsearch $::content_suffix_list $suffix]
        if { $index >= 0 } {
            set function [lindex [lindex $::content_handler_list $index] 3]
            if {[catch [$function $chan "$filename.$suffix" $head_req $last_modified] error]} {
                print_html_error $chan "$function '$filename.$suffix' '$head_req' '$last_modified'" $error
                puts $chan [html_end]
            }
        } else {
            print_html_header_404 $chan
        }
    } else {
        if {$head_req == 1} {
            catch {close $chan}
            return
        }

        if {$::url_prefix != ""} {
            if {[regexp "^/$::url_prefix" $path] == 1} {
                set path [string range $path [string length $::url_prefix] end]
            }
        }

        if {[regexp {^/([-_A-Za-z0-9]*)(.*)} $path dummy action part] == 1 &&
            [info procs "action_$action"] == "action_$action"} {
                   print_html_header_200 $chan "text/html; charset=iso-8859-1" ""
                   set env [parse_post $post_data]
                   if {[catch {eval {::action_$action $chan $part $env}}]} {
                      print_html_error $chan "action_$action '$part' '$env'" $::errorInfo
                      puts $chan [html_end]
                   }
        } else {
            print_html_header_404 $chan
        }
    }
    catch {flush $chan}
    catch {close $chan}
}

proc readconfig {} {
    global userpass
    global source_dir
    global http_port
    global url_prefix
    global description_hover

    set user ""
    set pass ""
    set http_port 80
    set url_prefix ""
    set default_theme "technophobe"
    set description_hover 1
    if {[catch {set fd [open "$source_dir/tivoweb.cfg" "r"]}]} {
        puts "Error opening configuration file 'tivoweb.cfg'"
    } else {
        set line [gets $fd]
        while { ![eof $fd] } {
            if {[regexp -nocase {^([^ ]+) *= *(.*)$} $line dummy varname value] == 1} {
                set varname [string tolower $varname]
                set value [string trim $value "\"'"]
                if {[string compare "username" $varname] == 0} {
                    set user $value
                }
                if {[string compare "password" $varname] == 0} {
                    set pass $value
                }
                if {[string compare "port" $varname] == 0} {
                    set http_port $value
                }
                if {[string compare "prefix" $varname] == 0} {
                    set url_prefix $value
                    if {[string range $url_prefix end end] != "/"} {
                        append url_prefix "/"
                    }
                    if {[string index $url_prefix 0] == "/"} {
                        set url_prefix [string range $url_prefix 1 end]
                    }
                }
                if {[string compare "theme" $varname] == 0} {
                    regsub ".css$" $value {} value
                    if { $value != "" } {
                        catch { set fd2 [open "$source_dir/$value.css" "r"] }
                        if { $fd2 != "" } {
                            set default_theme $value
                            close $fd2
                        }
                    }
                }
                if {[string compare "descriptionhover" $varname] == 0} {
                    set value [string tolower $value]
                    if {[string compare "yes" $value] == 0 ||
                        [string compare "on" $value] == 0 ||
                        [string compare "true" $value] == 0 ||
                        $value == "1"} {
                        set description_hover 1
                    } elseif {[string compare "no" $value] == 0 ||
                              [string compare "off" $value] == 0 ||
                              [string compare "false" $value] == 0 ||
                              $value == "0"} {
                        set description_hover 0
                    }
                }
            }
            set line [gets $fd]
        }
        close $fd
    }

    if {[string length $user] > 0 && [string length $pass] > 0} {
        set userpass "$user:$pass"
    } else {
        set userpass ""
    }
    if {$TT_HTTPD::STYLE == "" } {
        set TT_HTTPD::STYLE $default_theme
    }

}

proc get_tzoffset {mfstz dst} {
   if { $mfstz <= 0 } {
      set tz $mfstz
   } else {
      set tzlist "-5 -6 -7 -8 -9 -10 0 1 2 3 4 5 6 7 8 9 10 11 12 -1 -2 -3 -4 -11 -12"
      set tz [lindex $tzlist [expr $mfstz - 1]]
   }
   if { $dst == 2 } {
      set date [clock format [clock seconds] -format "%1d %w %1m %1H %1M"]
      scan $date "%d %d %d %d %d" dom dow month hour min
      if {$month > 4 && $month < 10} {
         set dlsval 1
      } elseif {$month == 4 && $dom > 7} {
         set dlsval 1
      } elseif {$month == 4 && $dom <= 7 && $dow == 0 && $hour >= 2} {
         set dlsval 1
      } elseif {$month == 4 && $dom <= 7 && $dow != 0 && ($dom-$dow > 0)} {
         set dlsval 1
      } elseif {$month == 10 && $dom < 25} {
         set dlsval 1
      } elseif {$month == 10 && $dom >= 25 && $dow == 0 && $hour < 2} {
         set dlsval 1
      } elseif {$month == 10 && $dom >= 25 && $dow != 0 && ($dom-24-$dow < 1) } {
         set dlsval 1
      } else {
         set dlsval 0
      }
      if {$dlsval == 1} {
         return [expr ($tz+1)*60*60]
      } else {
         return [expr $tz*60*60]
      }
   } else {
      return [expr $tz*60*60]
   }
}

global db
global startuptime
global module_list
global reload
global tivoswversion
global content_suffix_list
global content_handler_list
global version3
global dtivo

set dbPoolSize [expr 300 * 1024]
set startuptime [clock seconds]
set db [dbopen $dbPoolSize]
set module_list ""
set content_list ""
set content_suffix_list ""
set content_handler_list ""
set reload 0

RetryTransaction {
   set swsystem [db $db open /SwSystem/ACTIVE]
   set tivoswversion [dbobj $swsystem get Name]
   set setup [db $db open /Setup]

   if { [PrefixMatches "3.0" $tivoswversion] } {
      set version3 1
   } else {
      set version3 0
   }
   if { [PrefixMatches "2.5.2" $tivoswversion] } {
      set dtivo 1
   } else {
      set dtivo 0
   }

   if {$::version3} {
      set lconfig  [db $db open /State/LocationConfig]
      set setuptz [dbobj $lconfig get TimeZoneOld]
      set daylightsavings [dbobj $lconfig get DaylightSavingsPolicy]
   } else {
      set setuptz [dbobj $setup get TimeZone]
      set daylightsavings [dbobj $setup get DaylightSavingsPolicy]
   } 
   set setuptz [defaultval 0 $setuptz]
   set daylightsavings [defaultval 2 $daylightsavings]
}
set tzoffset [get_tzoffset $setuptz $daylightsavings]

register_content_handler "js" "text/javascript" "$source_dir" 0 serve_file
register_content_handler "css" "text/css" "$source_dir" 0 serve_file
register_content_handler "png" "image/png" "$source_dir/images" 1 serve_image

set sock ""

while { $reload != "quit" } {

	readconfig
  if { $sock != "" } { close $sock }
  set sock [socket -server ::session $http_port]


	#set tcl_traceExec 1

	puts "Loading modules..."
  set module_list ""
	set modules [glob "$source_dir/modules/*.itcl"]
	set modules [lsort $modules]
	foreach module $modules {
		set errorCode ""
   	set errorInfo ""
   	puts [file rootname [file tail $module]]
	shaketcl
   	if {[catch { source $module }]} {
   	 puts "$errorCode $errorInfo"
    }
	}


	make_menu

	puts "Accepting Connections"
	vwait reload
}

puts "$TT_HTTPD::EXIT_MESSAGE"



